# Averist
# Copyright (C) 1999-2000 Henrik Edlund <henrik@edlund.org>
# http://www.edlund.org/hacks/averist/
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA 02111-1307 USA.
#
# If you like Averist, please express your satisfaction with a donation: send
# me what you feel Averist has been worth to you. If you are glad that I
# developed Averist and distribute it as free software, rather than following
# the obstructive and antisocial practices typical of software developers,
# reward me. If you would like me to develop more free software, contribute.



# login(set) -> (username, session)
# login(set, query) -> (username, session)
# logout() -> ()
# vacuum() -> ()



package Averist;

use strict;
no strict 'refs';
use Fcntl qw(:DEFAULT :flock);
use Digest::MD5 ();
use Digest::SHA1 ();

use vars qw($Product_Name $Product_Version);
$Product_Name = 'Averist';
$Product_Version = '0.9.0';



#################  Do NOT edit anything outside the Configuration if you don't
# Configuration #  know what you are doing. If you are planning to tweak then
#################  please read the License first.

use vars qw(%Config);
%Config = (
	   'events' => {
	       'authentication' => {
		   'remote' => 'cgi',  # cgi
		   'local'  => 'file', # file | ldap | sql | tie
	       },
	       'reauthentication' => {
		   'remote' => 'cgi', # cgi | cookie
		   'local'  => 'sql', # none | sql | tie
	       },
	   },
	   'methods' => {
	       'authentication' => {
		   'remote' => {
		       'cgi' => {
			   'username' => 'username',
			   'password' => 'password',
		       },
		   },
		   'local' => {
		       'file' => {
			   'filename' => '/etc/httpd/conf/auth.conf',
			   'hash'     => 'crypt', # auto | crypt | md5 |
						  # md5_base64 | md5_hex |
						  # none | sha1 |
						  # sha1_base64 | sha1_hex
		       },
		       'ldap' => {
			   'server' => {
			       'host' => 'localhost',
			       'port' => '',
			   },
			   'bind' => {
			       'dn'       => 'cn=Manager, o=Name, c=SE',
			       'password' => 'secret',
			       'sasl'     => 'no', # yes | no
			   },
			   'search' => {
			       'base'     => 'ou=Staff, o=Name, c=SE',
			       'scope'    => 'sub', # base | one | sub
			       'username' => 'uid',
			       'password' => 'userpassword',
			   },
			   'hash' => 'crypt', # auto | crypt | md5 |
					      # md5_base64 | md5_hex | none |
					      # sha1 | sha1_base64 | sha1_hex
		       },
		       'sql' => {
			   'abstraction' => 'PostgreSQL', # MySQL | MSSQL |
							  # Oracle |
							  # PostgreSQL
			   'database' => {
			       'host'     => '',
			       'port'     => '',
			       'name'     => 'averist',
			       'user'     => 'henrik',
			       'password' => '',
			   },
			   'tables' => {
			       'users' => 'users',
			   },
			   'fields' => {
			       'users' => {
				   'username' => 'username',
				   'password' => 'passphrase',
			       },
			   },
			   'hash' => 'crypt', # auto | crypt | md5 |
					      # md5_base64 | md5_hex | none |
					      # sha1 | sha1_base64 | sha1_hex
		       },
		       'tie' => {
			   'interface' => 'DB_File',  # DB_File | GDBM_File |
						      # NDBM_File |
						      # ODBM_File | SDBM_File
			   'file'      => 'users.db',
			   'lock'      => ['users.db', undef],
			   'hash'      => 'crypt',    # auto | crypt | md5 |
						      # md5_base64 | md5_hex |
						      # none | sha1 |
						      # sha1_base64 | sha1_hex
		       },
		   },
	       },
	       'reauthentication' => {
		   'remote' => {
		       'cgi' => {
			   'username' => 'username',
			   'session'  => 'session',
		       },
		       'cookie' => {
			   'username' => 'username',
			   'session'  => 'session',
		       },
		   },
		   'local' => {
		       'sql' => {
			   'abstraction' => 'PostgreSQL', # MySQL | MSSQL |
							  # Oracle |
							  # PostgreSQL
			   'database' => {
			       'host'     => '',
			       'port'     => '',
			       'name'     => 'averist',
			       'user'     => 'henrik',
			       'password' => '',
			   },
			   'tables' => {
			       'sessions' => 'sessions',
			   },
			   'fields' => {
			       'sessions' => {
				   'username' => 'username',
				   'address'  => 'address',
				   'session'  => 'key',
				   'time'     => 'point',
			       },
			   },
		       },
		       'tie' => {
			   'interface' => 'DB_File',     # DB_File |
							 # GDBM_File | 
							 # NDBM_File |
							 # ODBM_File |
							 # SDBM_File
			   'file'      => 'sessions.db',
			   'lock'      => ['sessions.db', undef],
		       },
		   },
	       },
	   },
	   'hash' => {
	       'session' => 'sha1_hex', # md5_hex | sha1_hex
	       'auto'    => {
		   'crypt' => 'crypt',
		   'MD5'   => 'md5_base64',
		   'SHA'   => 'sha1_base64',
	       },
	   },
	   'time' => {
	       'session' => 5 * 60,       # 5 minutes
	       'vacuum'  => 24 * 60 * 60, # 1 day
	   },
);

######################
# Configuration Ends #
######################



if ($Config{'events'}{'authentication'}{'local'} eq 'ldap') {
    require Net::LDAP;
    if ($Config{'methods'}{'authentication'}{'local'}{'ldap'}{'bind'}{'sasl'}
	eq 'yes') {
	require Authen::SASL;
    }
}

use vars qw($_Sql_A_L);
$_Sql_A_L = $Config{'methods'}{'authentication'}{'local'}{'sql'}
{'abstraction'};
if ($Config{'events'}{'authentication'}{'local'} eq 'sql') {
    require "$_Sql_A_L\.pm";
}

use vars qw($_Sql_R_L);
$_Sql_R_L = $Config{'methods'}{'reauthentication'}{'local'}{'sql'}
{'abstraction'};
if ($Config{'events'}{'reauthentication'}{'local'} eq 'sql') {
    require "$_Sql_R_L\.pm";
}

if ($Config{'events'}{'authentication'}{'local'} eq 'tie') {
    $_ = $Config{'methods'}{'authentication'}{'local'}{'tie'}{'interface'};
    require "$_\.pm";
}

if ($Config{'events'}{'reauthentication'}{'local'} eq 'tie') {
    $_ = $Config{'methods'}{'reauthentication'}{'local'}{'tie'}{'interface'};
    require "$_\.pm";
}

use vars qw(%_Cgi_Query);

sub login ($;$) {
    (my $set) = @_;
    my $remote_username;
    my $remote_address;
    my $local_password;
    my $session;
    my $time;
    my $remote_session;
    my $local_session;
    my $local_time;
    my $remote_password;
    if (defined(%{$_[1]})) {
	%_Cgi_Query = %{$_[1]};
    }
    else {
	%_Cgi_Query = _parse_cgi_query();
    }
    if (($remote_username = _remote_username()) and
	($remote_address = _remote_address()) and
	($local_password = _local_password($remote_username))) {
	$session = _session($local_password);
	$time = time();
	if ((($remote_session = _remote_session()) and
	     ((($Config{'events'}{'reauthentication'}{'local'} eq 'none') and
	       ($remote_session eq $session)) or
	      (($Config{'events'}{'reauthentication'}{'local'} ne 'none') and
	       (($local_session, $local_time) =
		_local_session($remote_username, $remote_address)) and
	       ($remote_session eq $local_session) and
	       ($time <= ($local_time + $Config{'time'}{'session'}))))) or
	    (($remote_password = _remote_password()) and
	     (_verify_password($local_password, $remote_password)) and
	     ($remote_session = $session))) {
	    if ($set) {
		_set_session($remote_username,
			     $remote_address,
			     $session,
			     $time);
		return($remote_username, $session);
	    }
	    return($remote_username, $remote_session);
	}
	_set_session($remote_username, $remote_address, '*', 0);
	return($remote_username, undef);
    }
    return(undef, undef);
}

sub logout () {
    my $remote_username;
    my $remote_address;
    if (($remote_username = _remote_username()) and
	($remote_address = _remote_address())) {
	_set_session($remote_username, $remote_address, '*', 0);
    }
}

sub vacuum () {
    my $now;
    my $sql;
    my $conn;
    my $q1;
    my $file;
    my %session;
    my $lock;
    my $key;
    my $value;
    my $session;
    my $time;
    $now = time();
    if ($Config{'events'}{'reauthentication'}{'local'} eq 'sql') {
	$sql = $Config{'methods'}{'reauthentication'}{'local'}{'sql'};
	$conn = &{"${_Sql_R_L}::connect"}($$sql{'database'}{'host'},
					  $$sql{'database'}{'port'},
					  $$sql{'database'}{'name'},
					  $$sql{'database'}{'user'},
					  $$sql{'database'}{'password'});
	$q1 = "delete from $$sql{'tables'}{'sessions'} where " .
	    "$$sql{'fields'}{'sessions'}{'time'} < " .
		"$now-$Config{'time'}{'vacuum'}";
	&{"${_Sql_R_L}::query"}($conn, $q1);
    }
    elsif ($Config{'events'}{'reauthentication'}{'local'} eq 'tie') {
	$file =
	    $Config{'methods'}{'reauthentication'}{'local'}{'tie'}{'file'};
	tie(%session,
	    $Config{'methods'}{'reauthentication'}{'local'}{'tie'}
	    {'interface'},
	    $file,
	    O_CREAT|O_RDWR,
	    0666);
	die "Can't open $file for tie, stopped"
	    unless (tied(%session));
	$lock =
	    $Config{'methods'}{'reauthentication'}{'local'}{'tie'}{'lock'};
	if (defined($$lock[0])) {
	    die "Can't open $$lock[0] for lock, stopped"
		unless (sysopen(FILE1, $$lock[0], O_RDWR));
	    die "Can't lock $$lock[0], stopped"
		unless (flock(FILE1, LOCK_EX));
	}
	if (defined($$lock[1])) {
	    die "Can't open $$lock[1] for lock, stopped"
		unless (sysopen(FILE2, $$lock[1], O_RDWR));
	    die "Can't lock $$lock[1], stopped"
		unless (flock(FILE2, LOCK_EX));
	}
	while (($key, $value) = each(%session)) {
	    ($session, $time) = split(/\0/, $value);
	    if ($time < ($now - $Config{'time'}{'vacuum'})) {
		delete($session{$key});
	    }
	}
	untie(%session);
	if (defined($$lock[0])) {
	    flock(FILE1, LOCK_UN);
	    close(FILE1);
	}
	if (defined($$lock[1])) {
	    flock(FILE2, LOCK_UN);
	    close(FILE2);
	}
    }
    else {
	die "Invalid value in events->reauthentication->local, stopped";
    }
}

sub _remote_username () {
    my %parsed;
    if ($Config{'events'}{'reauthentication'}{'remote'} eq 'cgi') {
	if (defined($_Cgi_Query{$Config{'methods'}{'reauthentication'}
				{'remote'}{'cgi'}{'username'}})) {
	    return($_Cgi_Query{$Config{'methods'}{'reauthentication'}
			       {'remote'}{'cgi'}{'username'}});
	}
    }
    elsif ($Config{'events'}{'reauthentication'}{'remote'} eq 'cookie') {
	%parsed = _parse_cookies();
	if (defined($parsed{$Config{'methods'}{'reauthentication'}
			    {'remote'}{'cookie'}{'username'}})) {
	    return($parsed{$Config{'methods'}{'reauthentication'}{'remote'}
			   {'cookie'}{'username'}});
	}
    }
    else {
	die "Invalid value in events->reauthentication->remote, stopped";
    }
    if ($Config{'events'}{'authentication'}{'remote'} eq 'cgi') {
	if (defined($_Cgi_Query{$Config{'methods'}{'authentication'}{'remote'}
				{'cgi'}{'username'}})) {
	    return($_Cgi_Query{$Config{'methods'}{'authentication'}{'remote'}
			       {'cgi'}{'username'}});
	}
    }
    else {
	die "Invalid value in events->authentication->remote, stopped";
    }
    return(undef);
}

sub _remote_address () {
    if (defined($ENV{'REMOTE_ADDR'})) {
	return($ENV{'REMOTE_ADDR'});
    }
    return(undef);
}

sub _remote_session () {
    my %parsed;
    if ($Config{'events'}{'reauthentication'}{'remote'} eq 'cgi') {
	if (defined($_Cgi_Query{$Config{'methods'}{'reauthentication'}
				{'remote'}{'cgi'}{'session'}})) {
	    return($_Cgi_Query{$Config{'methods'}{'reauthentication'}
			       {'remote'}{'cgi'}{'session'}});
	}
    }
    elsif ($Config{'events'}{'reauthentication'}{'remote'} eq 'cookie') {
	%parsed = _parse_cookies();
	if (defined($parsed{$Config{'methods'}{'reauthentication'}{'remote'}
			    {'cookie'}{'session'}})) {
	    return($parsed{$Config{'methods'}{'reauthentication'}{'remote'}
			   {'cookie'}{'session'}});
	}
    }
    else {
	die "Invalid value in events->reauthentication->remote, stopped";
    }
    return(undef);
}

sub _remote_password () {
    if ($Config{'events'}{'authentication'}{'remote'} eq 'cgi') {
	if (defined($_Cgi_Query{$Config{'methods'}{'authentication'}{'remote'}
				{'cgi'}{'password'}})) {
	    return($_Cgi_Query{$Config{'methods'}{'authentication'}{'remote'}
			       {'cgi'}{'password'}});
	}
    }
    else {
	die "Invalid value in events->authentication->remote, stopped";
    }
    return(undef);
}

sub _local_session ($$) {
    (my $username, my $address) = @_;
    my $sql;
    my $conn;
    my $q1;
    my $file;
    my %session;
    my $lock;
    my $session;
    my $time;
    if ($Config{'events'}{'reauthentication'}{'local'} eq 'sql') {
	$sql = $Config{'methods'}{'reauthentication'}{'local'}{'sql'};
	$conn = &{"${_Sql_R_L}::connect"}($$sql{'database'}{'host'},
					  $$sql{'database'}{'port'},
					  $$sql{'database'}{'name'},
					  $$sql{'database'}{'user'},
					  $$sql{'database'}{'password'});
	$q1 = "select $$sql{'fields'}{'sessions'}{'session'}, " .
	    "$$sql{'fields'}{'sessions'}{'time'} from " .
		"$$sql{'tables'}{'sessions'} where " .
		    "$$sql{'fields'}{'sessions'}{'username'}=" .
			_quote($username) . " and " .
			    "$$sql{'fields'}{'sessions'}{'address'}=" .
				_quote($address);
	if (&{"${_Sql_R_L}::ntuples"}(&{"${_Sql_R_L}::query"}($conn,
							      $q1)) == 0) {
	    return(undef, undef);
	}
	else {
	    return(&{"${_Sql_R_L}::fetchrow"}(&{"${_Sql_R_L}::query"}($conn,
								      $q1)));
	}
    }
    elsif ($Config{'events'}{'reauthentication'}{'local'} eq 'tie') {
	$file =
	    $Config{'methods'}{'reauthentication'}{'local'}{'tie'}{'file'};
	tie(%session,
	    $Config{'methods'}{'reauthentication'}{'local'}{'tie'}
	    {'interface'},
	    $file,
	    O_CREAT|O_RDONLY,
	    0666);
	die "Can't open $file for tie, stopped"
	    unless (tied(%session));
	$lock =
	    $Config{'methods'}{'reauthentication'}{'local'}{'tie'}{'lock'};
	if (defined($$lock[0])) {
	    die "Can't open $$lock[0] for lock, stopped"
		unless (sysopen(FILE1, $$lock[0], O_RDONLY));
	    die "Can't lock $$lock[0], stopped"
		unless (flock(FILE1, LOCK_SH));
	}
	if (defined($$lock[1])) {
	    die "Can't open $$lock[1] for lock, stopped"
		unless (sysopen(FILE2, $$lock[1], O_RDONLY));
	    die "Can't lock $$lock[1], stopped"
		unless (flock(FILE2, LOCK_SH));
	}
	if (defined($session{"$username\0$address"})) {
	    ($session, $time) = split(/\0/, $session{"$username\0$address"});
	}
	elsif (defined($session{"$username\0$address\0"})) {
	    ($session, $time) = split(/\0/,
				      $session{"$username\0$address\0"});
	}
	untie(%session);
	if (defined($$lock[0])) {
	    flock(FILE1, LOCK_UN);
	    close(FILE1);
	}
	if (defined($$lock[1])) {
	    flock(FILE2, LOCK_UN);
	    close(FILE2);
	}
	return($session, time);
    }
    else {
	die "Invalid value in events->reauthentication->local, stopped";
    }
}

sub _local_password ($) {
    (my $username) = @_;
    my $file;
    my $read_username;
    my $read_password;
    my $ldap;
    my $sasl;
    my $result;
    my $sql;
    my $conn;
    my $q1;
    my %password;
    my $lock;
    my $password;
    if ($Config{'events'}{'authentication'}{'local'} eq 'file') {
	$file =
	    $Config{'methods'}{'authentication'}{'local'}{'file'}{'filename'};
	die "Can't open $file for reading, stopped"
	    unless (sysopen(FILE, $file, O_RDONLY));
	die "Can't lock $file, stopped"
	    unless (flock(FILE, LOCK_SH));
 	foreach (<FILE>) {
	    ($read_username, $read_password, undef) = split(/:/, $_, 3);
	    if ($username eq $read_username) {
		$read_password =~ s/\n//;
		flock(FILE, LOCK_UN);
		close(FILE);
		return($read_password);
	    }
	}
	flock(FILE, LOCK_UN);
	close(FILE);
	return(undef);
    }
    elsif ($Config{'events'}{'authentication'}{'local'} eq 'ldap') {
	$ldap = $Config{'methods'}{'authentication'}{'local'}{'ldap'};
	if ($$ldap{'server'}{'port'} ne '') {
	    $conn = Net::LDAP->new($$ldap{'server'}{'host'},
				   port => $$ldap{'server'}{'port'})
		or die "Can't connect to directory, stopped";
	}
	else {
	    $conn = Net::LDAP->new($$ldap{'server'}{'host'}) or
		die "Can't connect to directory, stopped";
	}
	if ($$ldap{'bind'}{'dn'} eq '') {
	    $result = $conn->bind(anonymous => 1);
	}
	elsif ($$ldap{'bind'}{'password'} eq '') {
	    $result = $conn->bind($$ldap{'bind'}{'dn'},
				  noauth => 1);
	}
	elsif ($$ldap{'bind'}{'sasl'} eq 'yes') {
	    $sasl = Authen::SASL->new('CRAM-MD5',
				      password => $$ldap{'bind'}{'password'});
	    $result = $conn->bind($$ldap{'bind'}{'dn'},
				  sasl => $sasl,
				  version => 3);
	}
	elsif ($$ldap{'bind'}{'sasl'} eq 'no') {
	    $result = $conn->bind($$ldap{'bind'}{'dn'},
				  password => $$ldap{'bind'}{'password'});
	}
	else {
	    die "Invalid value in " .
		"methods->authentication->local->ldap->bind->sasl, stopped";
	}
	if ($result->code) {
	    die "Can't bind to directory, stopped";
	}
	$result = $conn->search(base => $$ldap{'search'}{'base'},
				scope => $$ldap{'search'}{'scope'},
				filter =>
				"($$ldap{'search'}{'username'}=$username)",
				attrs => [$$ldap{'search'}{'password'}]);
	if ($result->count != 1) {
	    return(undef);
	}
	else {
	    return(($result->entry(0))->get_value($$ldap{'search'}
						  {'password'}));
	}
    }
    elsif ($Config{'events'}{'authentication'}{'local'} eq 'sql') {
	$sql = $Config{'methods'}{'authentication'}{'local'}{'sql'};
	$conn = &{"${_Sql_A_L}::connect"}($$sql{'database'}{'host'},
					  $$sql{'database'}{'port'},
					  $$sql{'database'}{'name'},
					  $$sql{'database'}{'user'},
					  $$sql{'database'}{'password'});
	$q1 = "select $$sql{'fields'}{'users'}{'password'} from " .
	    "$$sql{'tables'}{'users'} where " .
		"$$sql{'fields'}{'users'}{'username'}=" .
		    _quote($username);
	if (&{"${_Sql_A_L}::ntuples"}(&{"${_Sql_A_L}::query"}($conn,
							      $q1)) == 0) {
	    return(undef);
	}
	else {
	    return((&{"${_Sql_A_L}::fetchrow"}(&{"${_Sql_A_L}::query"}
					       ($conn, $q1)))[0]);
	}
    }
    elsif ($Config{'events'}{'authentication'}{'local'} eq 'tie') {
	$file = $Config{'methods'}{'authentication'}{'local'}{'tie'}{'file'};
	tie(%password,
	    $Config{'methods'}{'authentication'}{'local'}{'tie'}{'interface'},
	    $file,
	    O_RDONLY,
	    0666);
	die "Can't open $file for tie, stopped"
	    unless (tied(%password));
	$lock = $Config{'methods'}{'authentication'}{'local'}{'tie'}{'lock'};
	if (defined($$lock[0])) {
	    die "Can't open $$lock[0] for lock, stopped"
		unless (sysopen(FILE1, $$lock[0], O_RDONLY));
	    die "Can't lock $$lock[0], stopped"
		unless (flock(FILE1, LOCK_SH));
	}
	if (defined($$lock[1])) {
	    die "Can't open $$lock[1] for lock, stopped"
		unless (sysopen(FILE2, $$lock[1], O_RDONLY));
	    die "Can't lock $$lock[1], stopped"
		unless (flock(FILE2, LOCK_SH));
	}
	if (defined($password{$username})) {
	    $password = $password{$username};
	}
	elsif (defined($password{"$username\0"})) {
	    $password = $password{"$username\0"};
	    $password =~ s/\0$//;
	}
	untie(%password);
	if (defined($$lock[0])) {
	    flock(FILE1, LOCK_UN);
	    close(FILE1);
	}
	if (defined($$lock[1])) {
	    flock(FILE2, LOCK_UN);
	    close(FILE2);
	}
	return($password);
    }
    else {
	die "Invalid value in events->authentication->local, stopped";
    }
}    

sub _verify_password ($$) {
    (my $local_password, my $remote_password) = @_;
    my $hash;
    if ($Config{'events'}{'authentication'}{'local'} eq 'file') {
	$hash = $Config{'methods'}{'authentication'}{'local'}{'file'}{'hash'};
    }
    elsif ($Config{'events'}{'authentication'}{'local'} eq 'ldap') {
	$hash = $Config{'methods'}{'authentication'}{'local'}{'ldap'}{'hash'};
    }
    elsif ($Config{'events'}{'authentication'}{'local'} eq 'sql') {
	$hash = $Config{'methods'}{'authentication'}{'local'}{'sql'}{'hash'};
    }
    elsif ($Config{'events'}{'authentication'}{'local'} eq 'tie') {
	$hash = $Config{'methods'}{'authentication'}{'local'}{'tie'}{'hash'};
    }
    else {
	die "Invalid value in events->authentication->local, stopped";
    }
    if ($hash eq 'auto') {
	if ($local_password =~ /{(.+?)}(.+)/) {
	    if (defined($Config{'hash'}{'auto'}{$1})) {
		$hash = $Config{'hash'}{'auto'}{$1};
	    }
	    else {
		die "Missing hash type in hash->auto, stopped";
	    }
	    $local_password = $2;
	}
	else {
	    $hash = 'none';
	}
    }
    if ($hash eq 'crypt') {
	$local_password = substr($local_password, 0, 13);
	$remote_password = crypt($remote_password, $local_password);
    }
    elsif ($hash eq 'md5') {
	$local_password = substr($local_password, 0, 16);
	$remote_password = Digest::MD5::md5($remote_password);
    }
    elsif ($hash eq 'md5_hex') {
	$local_password = substr($local_password, 0, 32);
	$remote_password = Digest::MD5::md5_hex($remote_password);
    }
    elsif ($hash eq 'md5_base64') {
	$local_password = substr($local_password, 0, 22);
	$remote_password = Digest::MD5::md5_base64($remote_password);
    }
    elsif ($hash eq 'sha1') {
	$local_password = substr($local_password, 0, 20);
	$remote_password = Digest::SHA1::sha1($remote_password);
    }
    elsif ($hash eq 'sha1_hex') {
	$local_password = substr($local_password, 0, 40);
	$remote_password = Digest::SHA1::sha1_hex($remote_password);
    }
    elsif ($hash eq 'sha1_base64') {
	$local_password = substr($local_password, 0, 27);
	$remote_password = Digest::SHA1::sha1_base64($remote_password);
    }
    elsif ($hash ne 'none') {
	die "Invalid value in methods->authentication->local->" .
	    $Config{'events'}{'authentication'}{'local'} . "->hash, stopped";
    }
    if ($local_password eq $remote_password) {
	return(1);
    }
    else {
	return(0);
    }
}

sub _session ($) {
    (my $password) = @_;
    my $length = 128;
    my $session;
    if ($Config{'events'}{'reauthentication'}{'local'} ne 'none') {
	$password = ''; 
	while ($length-- > 0) {
	    $password .= chr(rand(256));
	}
    }
    if ($Config{'hash'}{'session'} eq 'md5_hex') {
	return(Digest::MD5::md5_hex($password));
    }
    elsif ($Config{'hash'}{'session'} eq 'sha1_hex') {
	return(Digest::SHA1::sha1_hex($password));
    }
    else {
	die "Invalid value in hash->session, stopped";
    }
}

sub _set_session ($$$$) {
    (my $username, my $address, my $session, my $time) = @_;
    if ($Config{'events'}{'reauthentication'}{'remote'} eq 'cookie') {
	_set_cookie($username, $session, $time);
    }
    if ($Config{'events'}{'reauthentication'}{'local'} eq 'sql') {
	_set_sql($username, $address, $session, $time);
    }
    elsif ($Config{'events'}{'reauthentication'}{'local'} eq 'tie') {
	_set_tie($username, $address, $session, $time);
    }
}

sub _set_cookie ($$$) {
    (my $username, my $session, my $time) = @_;
    my $cookie;
    my @weekdays;
    my @months;
    my $second;
    my $minute;
    my $hour;
    my $day;
    my $month;
    my $year;
    my $weekday;
    my $expire;
    $cookie = $Config{'methods'}{'reauthentication'}{'remote'}{'cookie'};
    @weekdays = ('Sunday',
		 'Monday',
		 'Tuesday',
		 'Wednesday',
		 'Thursday',
		 'Friday',
		 'Saturday');
    @months = ('Jan',
	       'Feb',
	       'Mar',
	       'Apr',
	       'May',
	       'Jun',
	       'Jul',
	       'Aug',
	       'Sep',
	       'Oct',
	       'Nov',
	       'Dec');
    ($second, $minute, $hour, $day, $month, $year, $weekday) =
	gmtime(($time + $Config{'time'}{'session'}));
    $expire = sprintf("%s, %.2u-%s-%u %.2u:%.2u:%.2u GMT",
		      $weekdays[$weekday],
		      $day,
		      $months[$month],
		      $year+1900,
		      $hour,
		      $minute,
		      $second);
    print "Set-Cookie: $$cookie{'username'}=$username";
    if ($time == 0) {
	print "; expires=$expire";
    }
    print "\n";
    print "Set-Cookie: $$cookie{'session'}=$session; expires=$expire\n";
}

sub _set_sql ($$$$) {
    (my $username, my $address, my $session, my $time) = @_;
    my $sql;
    my $conn;
    my $q1;
    $sql = $Config{'methods'}{'reauthentication'}{'local'}{'sql'};
    $conn = &{"${_Sql_R_L}::connect"}($$sql{'database'}{'host'},
				      $$sql{'database'}{'port'},
				      $$sql{'database'}{'name'},
				      $$sql{'database'}{'user'},
				      $$sql{'database'}{'password'});
    $q1 = "select * from $$sql{'tables'}{'sessions'} where " .
	"$$sql{'fields'}{'sessions'}{'username'}=" .
	    _quote($username) . " and " .
		"$$sql{'fields'}{'sessions'}{'address'}=" .
		    _quote($address);
    if (&{"${_Sql_R_L}::ntuples"}(&{"${_Sql_R_L}::query"}($conn, $q1)) == 0) {
	$q1 = "insert into $$sql{'tables'}{'sessions'} " .
	    "($$sql{'fields'}{'sessions'}{'username'}, " .
		"$$sql{'fields'}{'sessions'}{'address'}, " .
		    "$$sql{'fields'}{'sessions'}{'session'}, " .
			"$$sql{'fields'}{'sessions'}{'time'}) " .
			    "values (" . _quote($username) . ", " .
				_quote($address) . ", " .
				    _quote($session) . ", $time)";
    }
    else {
	$q1 = "update $$sql{'tables'}{'sessions'} set " .
	    "$$sql{'fields'}{'sessions'}{'session'}=" .
		_quote($session) . ", " .
		    "$$sql{'fields'}{'sessions'}{'time'}=$time where " .
			"$$sql{'fields'}{'sessions'}{'username'}=" .
			    _quote($username) . " and " .
				"$$sql{'fields'}{'sessions'}{'address'}=" .
				    _quote($address);
    }
    &{"${_Sql_R_L}::query"}($conn, $q1);
}

sub _set_tie ($$$$) {
    (my $username, my $address, my $session, my $time) = @_;
    my $file;
    my %session;
    my $lock;
    $file = $Config{'methods'}{'reauthentication'}{'local'}{'tie'}{'file'};
    tie(%session,
	$Config{'methods'}{'reauthentication'}{'local'}{'tie'}{'interface'},
	$file,
	O_CREAT|O_RDWR,
	0666);
    die "Can't open $file for tie, stopped"
	unless (tied(%session));
    $lock = $Config{'methods'}{'reauthentication'}{'local'}{'tie'}{'lock'};
    if (defined($$lock[0])) {
	die "Can't open $$lock[0] for lock, stopped"
	    unless (sysopen(FILE1, $$lock[0], O_RDWR));
	die "Can't lock $$lock[0], stopped"
	    unless (flock(FILE1, LOCK_EX));
    }
    if (defined($$lock[1])) {
	die "Can't open $$lock[1] for lock, stopped"
	    unless (sysopen(FILE2, $$lock[1], O_RDWR));
	die "Can't lock $$lock[1], stopped"
	    unless (flock(FILE2, LOCK_EX));
    }
    $session{"$username\0$address"} = "$session\0$time";
    untie(%session);
    if (defined($$lock[0])) {
	flock(FILE1, LOCK_UN);
	close(FILE1);
    }
    if (defined($$lock[1])) {
	flock(FILE2, LOCK_UN);
	close(FILE2);
    }
}

sub _quote ($) {
    (my $string) = @_;
    if (defined($string)) {
	$string =~ s/\'/\'\'/g;
	return "'$string'";
    }
    else {
	return "NULL";
    }
}

sub _parse_cgi_query () {
    my $in;
    my @in;
    my $key;
    my $val;
    my %in;
    if (defined($ENV{'REQUEST_METHOD'})) {
	if ($ENV{'REQUEST_METHOD'} eq 'GET') {
	    $in = $ENV{'QUERY_STRING'};
	}
	elsif ($ENV{'REQUEST_METHOD'} eq 'POST') {
	    if (defined($ENV{'CONTENT_LENGTH'})) {
		read(STDIN, $in, $ENV{'CONTENT_LENGTH'});
	    }
	    else {
		return();
	    }
	}
	else {
	    return();
	}
    }
    else {
	return();
    }
    @in = split(/[&;]+/, $in);
    foreach (@in) {
	s/\+/ /g;
	($key, $val) = split(/=/);
	$key =~ s/%(..)/pack("c", hex($1))/ge;
	$val =~ s/%(..)/pack("c", hex($1))/ge;
	$key =~ s/\0//g;
	$val =~ s/\0//g;
	$in{$key} = $val;
    }
    return(%in);
}

sub _parse_cookies () {
    my $in;
    my @in;
    my $key;
    my $val;
    my %in;
    if (defined($ENV{'HTTP_COOKIE'})) {
	$in = $ENV{'HTTP_COOKIE'};
    }
    else {
	return();
    }
    @in = split(/; /, $in);
    foreach (@in) {
	($key, $val) = split(/=/);
	$key =~ s/%(..)/pack("c", hex($1))/ge;
	$val =~ s/%(..)/pack("c", hex($1))/ge;
	$key =~ s/\0//g;
	$val =~ s/\0//g;
	$in{$key} = $val;
    }
    return(%in);
}

1;
